function Beamforming_demo()
% demonstrate the SBL_v3p12 code using a beamforming demo

% Environment parameters
c = 1500;       % speed of sound
f = 200;        % frequency
lambda = c/f;   % wavelength

% ULA-horizontal array configuration
Nsensor = 20;               % number of sensors
d = 1/2*lambda;             % intersensor spacing
q = (0:1:(Nsensor-1))';     % sensor numbering
xq = (q-(Nsensor-1)/2)*d;   % sensor locations

% sensor configuration structure
Sensor_s.Nsensor = Nsensor;
Sensor_s.lambda = lambda;
Sensor_s.d = d;
Sensor_s.q = q;
Sensor_s.xq = xq;

% SBL options structure for version 3
options = SBLSet();
options.convergence.error = 10^(-4);

% signal generation parameters
SNR = 0;

% total number of snapshots
Nsnapshot = 50;
   
% number of random repetitions to calculate the average performance 
Nsim = 1;

% range of angle space
thetalim = [-90 90];

theta_separation = 0.5;

% Bearing grid
theta = (thetalim(1):theta_separation:thetalim(2))';
Ntheta = length(theta);

% Design/steering matrix
sin_theta = sind(theta);
A = exp(-1i*2*pi/lambda*xq*sin_theta.')/sqrt(Nsensor);

% to set Nsource parameter
Signal_s = generate_signal(Sensor_s, Nsnapshot, SNR, 0);
options.Nsource = Signal_s.Nsources;

% multiple simulations
for isim = 1:Nsim
    figure;
    
    Signal_s = generate_signal(Sensor_s, Nsnapshot, SNR, isim);
    Ysignal = Signal_s.Ysignal;
    
    % run SBL
    display('Running SBL code');
    tic;
    [gamma, report] = SBL_v3p12( A, Ysignal, options );
    toc;

    % run CVX
    display('Running CVX code');
    mu = 70;    % regularization parameter
    tic;
    cvx_begin
        variable x_l1(Ntheta, Nsnapshot) complex;
        cvx_quiet(true)
        minimize( square_pos(norm(A * x_l1 - Ysignal, 'fro')) + mu * sum(norms(x_l1,2,2)) ) ;
    cvx_end
    toc;
    
    % plot abs(Xpost)
    subplot(2,2,1);
    imagesc(abs(squeeze(report.results.final_iteration.x_post)));
    title('SBL');
    subplot(2,2,3);
    imagesc(abs(x_l1));
    title('L1 - CVX');
    
    % plot gamma and amp^2
    subplot(2,2,2);
    plot(theta, gamma); hold on;
    plot(Signal_s.source_theta, 0, 'rx', 'Markersize', 10);
    legend('Gamma', 'True source location');
    title('SBL');
    grid on;
    subplot(2,2,4);
    plot(theta, abs(x_l1).^2); hold on;
    plot(Signal_s.source_theta, 0, 'rx', 'Markersize', 10);
    legend('X^2', 'True source location');
    title('L1 - CVX');
    grid on;

end

end


function Signal_s = generate_signal(Sensor_s, Nsnapshot, snr, seed)
% function to generate sensor observations

% sensor settings
Nsensor = Sensor_s.Nsensor;
lambda = Sensor_s.lambda;
xq = Sensor_s.xq;

% three DOA example
Nsources = 3;
source_theta = [-2; 3; 75];
source_amp = [4; 13; 10];

Ysignal = zeros(Nsensor,Nsnapshot);

% random number seed
rng(seed, 'twister');

% simulate source motion
for t = 1:Nsnapshot
    Xsource = source_amp.*exp(1i*2*pi*rand(Nsources,1));    % random phase
    
    % Represenation matrix (steering matrix)
    u = sind(source_theta);
    A = exp(-1i*2*pi/lambda*xq*u.')/sqrt(Nsensor);
    
    % Signal without noise
    Ysignal(:,t) = sum(A*diag(Xsource),2);
    
    % add noise to the signals
    rnl = 10^(-snr/20)*norm(Xsource);
    nwhite = complex(randn(Nsensor,1),randn(Nsensor,1))/sqrt(2*Nsensor);
    e = nwhite * rnl;	% error vector
    Ysignal(:,t) = Ysignal(:,t) + e;
end

% output signal structure
Signal_s.Nsnapshot = Nsnapshot;
Signal_s.Nsources = Nsources;
Signal_s.source_theta = source_theta;
Signal_s.Xsource = Xsource;
Signal_s.Ysignal = Ysignal;

end
